@@ -289,6 +289,77 @@ class TestRestController(PecanTestCase):
         assert r.status_int == 200
         assert r.body == b_(dumps(dict(items=ThingsController.data)))
 
+    def test_404_with_lookup(self):
+
+        class LookupController(RestController):
+
+            def __init__(self, _id):
+                self._id = _id
+
+            @expose()
+            def get_all(self):
+                return 'ID: %s' % self._id
+
+        class ThingsController(RestController):
+
+            @expose()
+            def _lookup(self, _id, *remainder):
+                return LookupController(_id), remainder
+
+        class RootController(object):
+            things = ThingsController()
+
+        # create the app
+        app = TestApp(make_app(RootController()))
+
+        # these should 404
+        for path in ('/things', '/things/'):
+            r = app.get(path, expect_errors=True)
+            assert r.status_int == 404
+
+        r = app.get('/things/foo')
+        assert r.status_int == 200
+        assert r.body == b_('ID: foo')
+
+    def test_getall_with_lookup(self):
+
+        class LookupController(RestController):
+
+            def __init__(self, _id):
+                self._id = _id
+
+            @expose()
+            def get_all(self):
+                return 'ID: %s' % self._id
+
+        class ThingsController(RestController):
+
+            data = ['zero', 'one', 'two', 'three']
+
+            @expose()
+            def _lookup(self, _id, *remainder):
+                return LookupController(_id), remainder
+
+            @expose('json')
+            def get_all(self):
+                return dict(items=self.data)
+
+        class RootController(object):
+            things = ThingsController()
+
+        # create the app
+        app = TestApp(make_app(RootController()))
+
+        # test get_all
+        for path in ('/things', '/things/'):
+            r = app.get(path)
+            assert r.status_int == 200
+            assert r.body == b_(dumps(dict(items=ThingsController.data)))
+
+        r = app.get('/things/foo')
+        assert r.status_int == 200
+        assert r.body == b_('ID: foo')
+
     def test_simple_nested_rest(self):
 
         class BarController(RestController):
@@ -670,6 +741,110 @@ class TestRestController(PecanTestCase):
             r = app.request('/things', method='RESET', status=404)
             assert r.status_int == 404
 
+    def test_nested_rest_with_missing_intermediate_id(self):
+
+        class BarsController(RestController):
+
+            data = [['zero-zero', 'zero-one'], ['one-zero', 'one-one']]
+
+            @expose('json')
+            def get_all(self, foo_id):
+                return dict(items=self.data[int(foo_id)])
+
+        class FoosController(RestController):
+
+            data = ['zero', 'one']
+
+            bars = BarsController()
+
+            @expose()
+            def get_one(self, id):
+                return self.data[int(id)]
+
+            @expose('json')
+            def get_all(self):
+                return dict(items=self.data)
+
+        class RootController(object):
+            foos = FoosController()
+
+        # create the app
+        app = TestApp(make_app(RootController()))
+
+        # test get_all
+        r = app.get('/foos')
+        assert r.status_int == 200
+        assert r.body == b_(dumps(dict(items=FoosController.data)))
+
+        # test nested get_all
+        r = app.get('/foos/1/bars')
+        assert r.status_int == 200
+        assert r.body == b_(dumps(dict(items=BarsController.data[1])))
+
+        r = app.get('/foos/bars', expect_errors=True)
+        assert r.status_int == 404
+
+    def test_custom_with_trailing_slash(self):
+
+        class CustomController(RestController):
+
+            _custom_actions = {
+                'detail': ['GET'],
+                'create': ['POST'],
+                'update': ['PUT'],
+                'remove': ['DELETE'],
+            }
+
+            @expose()
+            def detail(self):
+                return 'DETAIL'
+
+            @expose()
+            def create(self):
+                return 'CREATE'
+
+            @expose()
+            def update(self, id):
+                return id
+
+            @expose()
+            def remove(self, id):
+                return id
+
+        app = TestApp(make_app(CustomController()))
+
+        r = app.get('/detail')
+        assert r.status_int == 200
+        assert r.body == b_('DETAIL')
+
+        r = app.get('/detail/')
+        assert r.status_int == 200
+        assert r.body == b_('DETAIL')
+
+        r = app.post('/create')
+        assert r.status_int == 200
+        assert r.body == b_('CREATE')
+
+        r = app.post('/create/')
+        assert r.status_int == 200
+        assert r.body == b_('CREATE')
+
+        r = app.put('/update/123')
+        assert r.status_int == 200
+        assert r.body == b_('123')
+
+        r = app.put('/update/123/')
+        assert r.status_int == 200
+        assert r.body == b_('123')
+
+        r = app.delete('/remove/456')
+        assert r.status_int == 200
+        assert r.body == b_('456')
+
+        r = app.delete('/remove/456/')
+        assert r.status_int == 200
+        assert r.body == b_('456')
+
     def test_custom_delete(self):
 
         class OthersController(object):
@@ -1057,6 +1232,23 @@ class TestRestController(PecanTestCase):
         assert r.status_int == 200
         assert r.body == b_('POST-2')
 
+    def test_nested_rest_with_default(self):
+
+        class FooController(RestController):
+
+            @expose()
+            def _default(self, *remainder):
+                return "DEFAULT %s" % remainder
+
+        class RootController(RestController):
+            foo = FooController()
+
+        app = TestApp(make_app(RootController()))
+
+        r = app.get('/foo/missing')
+        assert r.status_int == 200
+        assert r.body == b_("DEFAULT missing")
+
     def test_dynamic_rest_lookup(self):
         class BarController(RestController):
             @expose()
